/*
** Copyright 2015 Centreon
**
** Licensed under the Apache License, Version 2.0 (the "License");
** you may not use this file except in compliance with the License.
** You may obtain a copy of the License at
**
**     http://www.apache.org/licenses/LICENSE-2.0
**
** Unless required by applicable law or agreed to in writing, software
** distributed under the License is distributed on an "AS IS" BASIS,
** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
** See the License for the specific language governing permissions and
** limitations under the License.
**
** For more information : contact@centreon.com
*/

#include <sstream>
#include "com/centreon/broker/io/events.hh"
#include "com/centreon/broker/multiplexing/publisher.hh"
#include "com/centreon/broker/dumper/internal.hh"
#include "com/centreon/broker/config/applier/state.hh"
#include "com/centreon/broker/database.hh"
#include "com/centreon/broker/database_query.hh"
#include "com/centreon/broker/dumper/db_dump.hh"
#include "com/centreon/broker/dumper/db_dump_committed.hh"
#include "com/centreon/broker/dumper/db_writer.hh"
#include "com/centreon/broker/dumper/entries/ba.hh"
#include "com/centreon/broker/dumper/entries/kpi.hh"
#include "com/centreon/broker/dumper/entries/host.hh"
#include "com/centreon/broker/dumper/entries/service.hh"
#include "com/centreon/broker/dumper/entries/boolean.hh"
#include "mapping.hh"
#include "com/centreon/broker/io/exceptions/shutdown.hh"
#include "com/centreon/broker/logging/logging.hh"

using namespace com::centreon::broker;
using namespace com::centreon::broker::dumper;

/**************************************
*                                     *
*           Public Methods            *
*                                     *
**************************************/

/**
 *  Constructor.
 *
 *  @param[in] db_cfg  Database configuration.
 */
db_writer::db_writer(database_config const& db_cfg)
  : _process_out(true), _db_cfg(db_cfg), _full_dump(false) {}

/**
 *  Destructor.
 */
db_writer::~db_writer() {}

/**
 *  Set processing flags.
 *
 *  @param[in] in  Set to true to process input events.
 *  @param[in] out Set to true to process output events.
 */
void db_writer::process(bool in, bool out) {
  _process_out = in || !out; // Only for immediate shutdown.
  return ;
}

/**
 *  db_writer cannot read. This method will throw.
 *
 *  @param[in] d         Unused.
 */
void db_writer::read(misc::shared_ptr<io::data>& d) {
  d.clear();
  throw (com::centreon::broker::exceptions::msg()
         << "influxdb: attempt to read from a db_writer stream");
}

/**
 *  Write events.
 *
 *  @param[in] d  Event to write.
 */
unsigned int db_writer::write(misc::shared_ptr<io::data> const& d) {
  // Check that processing is enabled.
  if (!_process_out)
    throw (io::exceptions::shutdown(true, true)
             << "db_writer stream is shutdown");

  if (!d.isNull()) {
    if (d->type() ==
        io::events::data_type<io::events::dumper, dumper::de_db_dump>::value) {
      db_dump const& dbd(d.ref_as<db_dump const>());
      if (dbd.poller_id
          == config::applier::state::instance().get_instance_id()) {
        if (dbd.commit)
          _commit();
        else
          _full_dump = dbd.full;
        _bas.clear();
        _booleans.clear();
        _kpis.clear();
        _hosts.clear();
        _services.clear();
        multiplexing::publisher pblsh;
        misc::shared_ptr<db_dump_committed> dbc(new db_dump_committed);
        pblsh.write(dbc);
      }
    }
    else if (d->type() ==
             io::events::data_type<io::events::dumper, dumper::de_entries_ba>::value) {
      entries::ba const& b(d.ref_as<entries::ba const>());
      if (b.poller_id == config::applier::state::instance().get_instance_id())
        _bas.push_back(b);
    }
    else if (d->type() ==
             io::events::data_type<io::events::dumper, dumper::de_entries_boolean>::value) {
      entries::boolean const& b(d.ref_as<entries::boolean const>());
      if (b.poller_id == config::applier::state::instance().get_instance_id())
        _booleans.push_back(b);
    }
    else if (d->type() ==
             io::events::data_type<io::events::dumper, dumper::de_entries_kpi>::value) {
      entries::kpi const& k(d.ref_as<entries::kpi const>());
      if (k.poller_id == config::applier::state::instance().get_instance_id())
        _kpis.push_back(k);
    }
    else if (d->type() ==
             io::events::data_type<io::events::dumper, dumper::de_entries_host>::value) {
      entries::host const& h(d.ref_as<entries::host const>());
      if (h.poller_id == config::applier::state::instance().get_instance_id())
        _hosts.push_back(h);
    }
    else if (d->type() ==
             io::events::data_type<io::events::dumper, dumper::de_entries_service>::value) {
      entries::service const& s(d.ref_as<entries::service const>());
      if (s.poller_id == config::applier::state::instance().get_instance_id())
        _services.push_back(s);
    }
  }
  return (1);
}

/**************************************
*                                     *
*           Private Methods           *
*                                     *
**************************************/

/**
 *  Commit configuration to database.
 */
void db_writer::_commit() {
  // Open DB connection.
  database db(_db_cfg);

  // Clean database if necessary.
  if (_full_dump) {
    {
      database_query q(db);
      q.run_query("DELETE FROM mod_bam_kpi");
    }
    {
      database_query q(db);
      q.run_query("DELETE FROM mod_bam");
    }
    {
      database_query q(db);
      q.run_query("DELETE FROM mod_bam_boolean");
    }
    {
      database_query q(db);
      q.run_query("DELETE FROM service WHERE service_description like 'ba_%'");
    }
  }

  // Prepare BA queries.
  database_query ba_insert(db);
  database_query ba_update(db);
  database_query ba_delete(db);
  {
    std::set<std::string> ids;
    ids.insert("ba_id");
    _prepare_insert<entries::ba>(ba_insert);
    _prepare_update<entries::ba>(ba_update, ids);
    _prepare_delete<entries::ba>(ba_delete, ids);
  }

  // Prepare boolean rules queries.
  database_query boolean_insert(db);
  database_query boolean_update(db);
  database_query boolean_delete(db);
  {
    std::set<std::string> ids;
    ids.insert("boolean_id");
    _prepare_insert<entries::boolean>(boolean_insert);
    _prepare_update<entries::boolean>(boolean_update, ids);
    _prepare_delete<entries::boolean>(boolean_delete, ids);
  }

  // Prepare KPI queries.
  database_query kpi_insert(db);
  database_query kpi_update(db);
  database_query kpi_delete(db);
  {
    std::set<std::string> ids;
    ids.insert("kpi_id");
    _prepare_insert<entries::kpi>(kpi_insert);
    _prepare_update<entries::kpi>(kpi_update, ids);
    _prepare_delete<entries::kpi>(kpi_delete, ids);
  }

  // Prepare host queries.
  database_query host_insert(db);
  database_query host_update(db);
  database_query host_delete(db);
  {
    std::set<std::string> ids;
    ids.insert("host_id");
    _prepare_insert<entries::host>(host_insert);
    _prepare_update<entries::host>(host_update, ids);
    _prepare_delete<entries::host>(host_delete, ids);
  }

  // Prepare service queries.
  database_query service_insert(db);
  database_query service_update(db);
  database_query service_delete(db);
  {
    std::set<std::string> ids;
    ids.insert("service_id");
    _prepare_insert<entries::service>(service_insert);
    _prepare_update<entries::service>(service_update, ids);
    _prepare_delete<entries::service>(service_delete, ids);
  }

  // Prepare relation table queries.
  database_query poller_ba_relation_insert(db);
  database_query host_service_relation_insert(db);
  {
    poller_ba_relation_insert.prepare(
      "INSERT INTO mod_bam_poller_relations"
      "  (ba_id, poller_id)"
      "  VALUES(:ba_id, :poller_id)",
      "db_writer: could not prepare insertion query");
  }
  {
    host_service_relation_insert.prepare(
      "INSERT INTO host_service_relation"
      "  (host_host_id, service_service_id)"
      "  VALUES(:host_id, :service_id)",
      "db_writer: could not prepare insertion query");
  }

  // Process all BAs.
  for (std::list<entries::ba>::const_iterator it(_bas.begin()), end(_bas.end());
       it != end;
       ++it) {
    // INSERT / UPDATE.
    if (it->enable) {
      logging::debug(logging::medium)
        << "db_dumper: updating BA " << it->ba_id << " ('" << it->name
        << "')";
      _fill_query(ba_update, *it);
      ba_update.run_statement();
      if (!ba_update.num_rows_affected()) {
        logging::debug(logging::medium)
          << "db_dumper: inserting BA " << it->ba_id << " ('"
          << it->name << "')";
        _fill_query(ba_insert, *it);
        ba_insert.run_statement();
      }
      std::ostringstream query;
      query << "UPDATE mod_bam SET activate='1' WHERE ba_id="
            << it->ba_id;
      database_query q(db);
      q.run_query(query.str().c_str());
      poller_ba_relation_insert.bind_value(":poller_id", it->poller_id);
      poller_ba_relation_insert.bind_value(":ba_id", it->ba_id);
      try {
        poller_ba_relation_insert.run_statement();
      } catch (std::exception const&) {}
    }
    // DELETE.
    else {
      logging::debug(logging::medium)
        << "db_dumper: deleting BA " << it->ba_id << " ('"
        << it->name << "')";
      ba_delete.bind_value(":ba_id", it->ba_id);
      ba_delete.run_statement();
    }
  }

  // Process all boolean rules.
  for (std::list<entries::boolean>::const_iterator it(_booleans.begin()), end(_booleans.end());
       it != end;
       ++it) {
    // INSERT / UPDATE.
    if (it->enable) {
      logging::debug(logging::medium)
        << "db_dumper: updating BA " << it->boolean_id << " ('" << it->name
        << "')";
      _fill_query(boolean_update, *it);
      boolean_update.run_statement();
      if (!boolean_update.num_rows_affected()) {
        logging::debug(logging::medium)
          << "db_dumper: inserting BA " << it->boolean_id << " ('"
          << it->name << "')";
        _fill_query(boolean_insert, *it);
        boolean_insert.run_statement();
      }
      std::ostringstream query;
      query << "UPDATE mod_bam_boolean SET activate='1' WHERE boolean_id="
            << it->boolean_id;
      database_query q(db);
      q.run_query(query.str().c_str());
    }
    // DELETE.
    else {
      logging::debug(logging::medium)
        << "db_dumper: deleting BA " << it->boolean_id << " ('"
        << it->name << "')";
      boolean_delete.bind_value(":boolean_id", it->boolean_id);
      boolean_delete.run_statement();
    }
  }

  // Process all KPIs.
  for (std::list<entries::kpi>::const_iterator
         it(_kpis.begin()),
         end(_kpis.end());
       it != end;
       ++it) {
    // INSERT / UPDATE.
    if (it->enable) {
      logging::debug(logging::medium)
        << "db_dumper: updating KPI " << it->kpi_id;
      _fill_query(kpi_update, *it);
      kpi_update.run_statement();
      if (!kpi_update.num_rows_affected()) {
        logging::debug(logging::medium)
          << "db_dumper: inserting KPI " << it->kpi_id;
        _fill_query(kpi_insert, *it);
        kpi_insert.run_statement();
      }
      std::ostringstream query;
      query << "UPDATE mod_bam_kpi SET activate='1', config_type='1'"
               " WHERE kpi_id=" << it->kpi_id;
      database_query q(db);
      q.run_query(query.str().c_str());
    }
    // DELETE.
    else {
      logging::debug(logging::medium)
        << "db_dumper: deleting KPI " << it->kpi_id;
      kpi_delete.bind_value(":kpi_id", it->kpi_id);
      kpi_delete.run_statement();
    }
  }

  // Process all hosts.
  for (std::list<entries::host>::const_iterator
         it(_hosts.begin()),
         end(_hosts.end());
       it != end;
       ++it) {
    // INSERT / UPDATE.
    if (it->enable) {
      logging::debug(logging::medium)
        << "db_dumper: updating host " << it->host_id;
      _fill_query(host_update, *it);
      host_update.run_statement();
      if (!host_update.num_rows_affected()) {
        logging::debug(logging::medium)
          << "db_dumper: inserting host " << it->host_id;
        _fill_query(host_insert, *it);
        host_insert.run_statement();
      }
    }
    // DELETE.
    else {
      logging::debug(logging::medium)
        << "db_dumper: deleting host " << it->host_id;
      host_delete.bind_value(":host_id", it->host_id);
      host_delete.run_statement();
    }
  }

  // Process all services
  for (std::list<entries::service>::const_iterator
         it(_services.begin()),
         end(_services.end());
       it != end;
       ++it) {
    // INSERT / UPDATE.
    if (it->enable) {
      logging::debug(logging::medium)
        << "db_dumper: updating service " << it->service_id;
      _fill_query(service_update, *it);
      service_update.run_statement();
      if (!service_update.num_rows_affected()) {
        logging::debug(logging::medium)
          << "db_dumper: inserting service " << it->service_id;
        _fill_query(service_insert, *it);
        service_insert.run_statement();
      }
      host_service_relation_insert.bind_value(":host_id", it->host_id);
      host_service_relation_insert.bind_value(":service_id", it->service_id);
      try {
        host_service_relation_insert.run_statement();
      } catch (std::exception const&) {}
    }
    // DELETE.
    else {
      logging::debug(logging::medium)
        << "db_dumper: deleting service " << it->service_id;
      service_delete.bind_value(":service_id", it->service_id);
      service_delete.run_statement();
    }
  }

  return ;
}

/**
 *  Prepare an insert statement for later execution.
 *
 *  @param[out] st  Query object.
 */
template <typename T>
void db_writer::_prepare_insert(database_query& st) {
  // Build query string.
  std::string query;
  query = "INSERT INTO ";
  query.append(mapped_type<T>::table);
  query.append(" (");
  for (unsigned int i = 0; mapped_type<T>::members[i].type; ++i)
    if (mapped_type<T>::members[i].name) {
      query.append(mapped_type<T>::members[i].name);
      query.append(", ");
    }

  query.resize(query.size() - 2);
  query.append(") VALUES(");
  for (unsigned int i = 0; mapped_type<T>::members[i].type; ++i)
    if (mapped_type<T>::members[i].name) {
      query.append(":");
      query.append(mapped_type<T>::members[i].name);
      query.append(", ");
    }
  query.resize(query.size() - 2);
  query.append(")");

  // Prepare statement.
  st.prepare(query, "db_writer: could not prepare insertion query");

  return ;
}

/**
 *  Prepare an update statement for later execution.
 *
 *  @param[out] st  Query object.
 *  @param[in]  id  List of fields that form an UNIQUE.
 */
template <typename T>
void db_writer::_prepare_update(
                  database_query& st,
                  std::set<std::string> const& id) {
  // Build query string.
  std::string query;
  query = "UPDATE ";
  query.append(mapped_type<T>::table);
  query.append(" SET ");
  for (unsigned int i = 0; mapped_type<T>::members[i].type; ++i)
    if (mapped_type<T>::members[i].name) {
      bool found(id.find(mapped_type<T>::members[i].name) !=  id.end());
      if (!found) {
        query.append(mapped_type<T>::members[i].name);
        query.append("= :");
        query.append(mapped_type<T>::members[i].name);
        query.append(", ");
      }
    }
  query.resize(query.size() - 2);
  query.append(" WHERE ");
  for (std::set<std::string>::const_iterator
         it(id.begin()),
         end(id.end());
       it != end;
       ++it) {
    query.append("COALESCE(");
    query.append(*it);
    query.append(", -1)=COALESCE(:");
    query.append(*it);
    query.append(", -1)");
    query.append(" AND ");
  }
  query.resize(query.size() - 5);

  // Prepare statement.
  st.prepare(query, "db_writer: could not prepare update query");

  return ;
}

/**
 *  Prepare a deletion query.
 *
 *  @param[out] st  Query object.
 *  @param[in]  id  List of fields that form an UNIQUE.
 */
template <typename T>
void db_writer::_prepare_delete(
                  database_query& st,
                  std::set<std::string> const& id) {
  // Build query string.
  std::string query;
  query = "DELETE FROM ";
  query.append(mapped_type<T>::table);
  query.append(" WHERE ");
  for (std::set<std::string>::const_iterator
         it(id.begin()),
         end(id.end());
       it != end;
       ++it) {
    query.append("COALESCE(");
    query.append(*it);
    query.append(", -1)=COALESCE(:");
    query.append(*it);
    query.append(", -1)");
    query.append(" AND ");
  }
  query.resize(query.size() - 5);

  // Prepare statement.
  st.prepare(query, "db_writer: could not prepare deletion query");

  return ;
}

/**
 *  Get a boolean from an object.
 */
static void bind_boolean(
              QString const& field,
              bool val,
              database_query& q) {
  q.bind_value(field, QVariant(val));
  return ;
}

/**
 *  Get a double from an object.
 */
static void bind_double(
              QString const& field,
              double val,
              database_query& q) {
  q.bind_value(field, QVariant(val));
  return ;
}

/**
 *  Get an integer from an object.
 */
static void bind_integer(
              QString const& field,
              int val,
              database_query& q) {
  q.bind_value(field, QVariant(val));
  return ;
}

/**
 *  Get an integer that is null on zero.
 */
static void bind_integer_null_on_zero(
              QString const& field,
              int val,
              database_query& q) {
  // Not-NULL.
  if (val)
    q.bind_value(field, QVariant(val));
  // NULL.
  else
    q.bind_value(field, QVariant(QVariant::Int));
  return ;
}

/**
 *  Get an integer that is null on -1.
 */
static void bind_integer_null_on_minus_one(
              QString const& field,
              int val,
              database_query& q) {
  // Not-NULL.
  if (val != -1)
    q.bind_value(field, QVariant(val));
  else
    q.bind_value(field, QVariant(QVariant::Int));
  return ;
}

/**
 *  Get a short from an object.
 */
static void bind_short(
              QString const& field,
              short val,
              database_query& q) {
  q.bind_value(field, QVariant(val));
  return ;
}

/**
 *  Get a string from an object.
 */
static void bind_string(
              QString const& field,
              QString const& val,
              database_query& q) {
  q.bind_value(field, QVariant(val.toStdString().c_str()));
  return ;
}

/**
 *  Get a string that might be null from an object.
 */
static void bind_string_null_on_empty(
              QString const& field,
              QString const& val,
              database_query& q) {
  // Not-NULL.
  if (!val.isEmpty())
    q.bind_value(field, QVariant(val.toStdString().c_str()));
  // NULL.
  else
    q.bind_value(field, QVariant(QVariant::String));
  return ;
}

/**
 *  Get a time_t from an object.
 */
static void bind_timet(
              QString const& field,
              time_t val,
              database_query& q) {
  q.bind_value(
      field,
      QVariant(static_cast<qlonglong>(val)));
  return ;
}

/**
 *  Get a time_t that is null on 0.
 */
static void bind_timet_null_on_zero(
              QString const& field,
              time_t val,
              database_query& q) {
  qlonglong vall(val);
  // Not-NULL.
  if (vall)
    q.bind_value(field, QVariant(vall));
  // NULL.
  else
    q.bind_value(field, QVariant(QVariant::LongLong));
  return ;
}

/**
 *  Get a time_t that is null on -1.
 */
static void bind_timet_null_on_minus_one(
              QString const& field,
              time_t val,
              database_query& q) {
  qlonglong vall(val);
  // Not-NULL.
  if (vall != -1)
    q.bind_value(field, QVariant(vall));
  // NULL.
  else
    q.bind_value(field, QVariant(QVariant::LongLong));
  return ;
}

/**
 *  Get an unsigned int from an object.
 */
static void bind_uint(
              QString const& field,
              unsigned int val,
              database_query& q) {
  q.bind_value(field, QVariant(val));
  return ;
}

/**
 *  Get an unsigned int that is null on zero.
 */
static void bind_uint_null_on_zero(
              QString const& field,
              unsigned int val,
              database_query& q) {
  // Not-NULL.
  if (val)
    q.bind_value(field, QVariant(val));
  // NULL
  else
    q.bind_value(field, QVariant(QVariant::Int));
  return ;
}

/**
 *  Get an unsigned int that is null on -1.
 */
static void bind_uint_null_on_minus_one(
              QString const& field,
              unsigned int val,
              database_query& q) {
  // Not-NULL.
  if (val != (unsigned int)-1)
    q.bind_value(field, QVariant(val));
  // NULL.
  else
    q.bind_value(field, QVariant(QVariant::Int));
  return ;
}

/**
 *  Bind an event to a query.
 *
 *  @param[in,out] st     The query.
 *  @param[in]     event  The event.
 */
template <typename T>
void db_writer::_fill_query(
                  database_query& st,
                  T const& event) {
  for (unsigned int i = 0; mapped_type<T>::members[i].type; ++i)
    if (mapped_type<T>::members[i].name) {
      QString field(":");
      field.append(mapped_type<T>::members[i].name);
      switch (mapped_type<T>::members[i].type) {
      case mapped_data<T>::BOOL:
        bind_boolean(field, event.*mapped_type<T>::members[i].member.b, st);
        break ;
      case mapped_data<T>::DOUBLE:
        bind_double(field, (event.*mapped_type<T>::members[i].member.d), st);
        break ;
      case mapped_data<T>::INT:
        switch (mapped_type<T>::members[i].null_on_value) {
        case NULL_ON_ZERO:
          bind_integer_null_on_zero(
            field,
           event.*mapped_type<T>::members[i].member.i,
            st);
          break ;
        case NULL_ON_MINUS_ONE:
          bind_integer_null_on_minus_one(
            field,
            event.*mapped_type<T>::members[i].member.i,
            st);
          break ;
        default:
          bind_integer(field, event.*mapped_type<T>::members[i].member.i, st);
        }
        break ;
      case mapped_data<T>::SHORT:
        bind_short(field, event.*mapped_type<T>::members[i].member.s, st);
        break ;
      case mapped_data<T>::STRING:
        if (mapped_type<T>::members[i].null_on_value
            == NULL_ON_ZERO)
          bind_string_null_on_empty(
            field,
            event.*mapped_type<T>::members[i].member.S,
            st);
        else
          bind_string(field, event.*mapped_type<T>::members[i].member.S, st);
        break ;
      case mapped_data<T>::TIMESTAMP:
        switch (mapped_type<T>::members[i].null_on_value) {
        case NULL_ON_ZERO:
          bind_timet_null_on_zero(
            field,
            event.*mapped_type<T>::members[i].member.t,
            st);
          break ;
        case NULL_ON_MINUS_ONE:
          bind_timet_null_on_minus_one(
            field,
            event.*mapped_type<T>::members[i].member.t,
            st);
          break ;
        default:
          bind_timet(field, event.*mapped_type<T>::members[i].member.t, st);
        }
        break ;
      case mapped_data<T>::UINT:
        switch (mapped_type<T>::members[i].null_on_value) {
        case NULL_ON_ZERO:
          bind_uint_null_on_zero(
            field,
            event.*mapped_type<T>::members[i].member.u,
            st);
          break ;
        case NULL_ON_MINUS_ONE:
          bind_uint_null_on_minus_one(
            field,
            event.*mapped_type<T>::members[i].member.u,
            st);
          break ;
        default :
          bind_uint(field, event.*mapped_type<T>::members[i].member.u, st);
        }
        break ;
      }
  }
}
