/*
** Copyright 2011-2014 Centreon
**
** Licensed under the Apache License, Version 2.0 (the "License");
** you may not use this file except in compliance with the License.
** You may obtain a copy of the License at
**
**     http://www.apache.org/licenses/LICENSE-2.0
**
** Unless required by applicable law or agreed to in writing, software
** distributed under the License is distributed on an "AS IS" BASIS,
** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
** See the License for the specific language governing permissions and
** limitations under the License.
**
** For more information : contact@centreon.com
*/

#include <QCoreApplication>
#include <QDir>
#include <QFile>
#include "com/centreon/broker/config/applier/init.hh"
#include "com/centreon/broker/file/stream.hh"
#include "com/centreon/broker/io/raw.hh"
#include "com/centreon/broker/ndo/internal.hh"
#include "com/centreon/broker/ndo/stream.hh"
#include "com/centreon/broker/neb/acknowledgement.hh"

using namespace com::centreon::broker;

#define TEMP_FILE "broker_ndo_stream_erased_input"

/**
 *  Check that NDO layer works properly with erased input.
 *
 *  @param[in] argc Argument count.
 *  @param[in] argv Argument values.
 *
 *  @return 0 on success.
 */
int main(int argc, char* argv[]) {
  // Qt core object.
  QCoreApplication app(argc, argv);

  // Initialization.
  config::applier::init();
  ndo::initialize();

  // File path.
  QString path;
  path = QDir::tempPath();
  path.append("/" TEMP_FILE);
  QFile::remove(path);

  // File stream.
  misc::shared_ptr<file::stream>
    fs(new file::stream(path.toStdString()));

  {
    // Buffer that will be written.
    misc::shared_ptr<io::raw> buffer(new io::raw);
    buffer->fill('\n', 8000);
    buffer->append(
      "42=21\n" \
      "36=12\n" \
      "999\n" \
      "\n");

    // Write buffer to file.
    fs->write(buffer);
  }

  // NDO stream.
  ndo::stream ns;
  ns.read_from(fs);
  ns.write_to(fs);

  // Sample object.
  misc::shared_ptr<neb::acknowledgement> ack(new neb::acknowledgement);
  ack->acknowledgement_type = 1;
  ack->author = "Matthieu Kermagoret";
  ack->comment = "This is a comment.";
  ack->entry_time = 123456789;
  ack->host_id = 42;
  ack->instance_id = 21;
  ack->is_sticky = false;
  ack->notify_contacts = true;
  ack->persistent_comment = true;
  ack->service_id = 84;
  ack->state = 2;

  // Write to NDO stream.
  ns.write(ack);

  // Now the check begins.
  int retval(0);
  misc::shared_ptr<io::data> d;
  ns.read(d);
  if (d.isNull() || (d->type() != ack->type()))
    retval = 1;
  else {
    misc::shared_ptr<neb::acknowledgement>
      readack(d.staticCast<neb::acknowledgement>());
    retval = ((ack->acknowledgement_type
               != readack->acknowledgement_type)
              || (ack->author != readack->author)
              || (ack->comment != readack->comment)
              || (ack->entry_time != readack->entry_time)
              || (ack->host_id != readack->host_id)
              || (ack->instance_id != readack->instance_id)
              || (ack->is_sticky != readack->is_sticky)
              || (ack->notify_contacts != readack->notify_contacts)
              || (ack->persistent_comment
                  != readack->persistent_comment)
              || (ack->service_id != readack->service_id)
              || (ack->state != readack->state));
  }

  // Erase temporary file.
  ns.read_from(misc::shared_ptr<io::stream>());
  ns.write_to(misc::shared_ptr<io::stream>());
  fs.clear();
  QFile::remove(path);

  return (retval);
}
